# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from ..parameters.with_parameters import WithParameters


class MeanFunction(WithParameters):
    """

    """
    def __init__(self, hyperparams):

        super(MeanFunction, self).__init__("Mean", hyperparams)

    def marginalize_mean(self, x):
        """

        :param x:
        :return:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dmu_dx(self, x):
        """

        :param x:
        :type x:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dmu_dtheta(self, x, trans=False):
        """

        :param x:
        :type x:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from mpl_toolkits.mplot3d import Axes3D

plt.style.use('ggplot')


def kernel(kernel_function, data=None, file_name=None):
    """
    Plot a kernel implementation

    :param kernel_function:
    :type kernel_function:
    :param data:
    :type data:
    :param file_name:
    :type file_name:
    :return:
    :rtype:
    """

    if data is None:
        data = np.linspace(-3.0, 3.0, 201)[:, None]
        index_mode = False
    else:
        index_mode = True

    middle_index = int(data.shape[0] * 0.5)
    middle_data = data[middle_index, :][None, :]

    plt.figure(figsize=(11, 4))
    plt.clf()
    plt.suptitle(kernel_function.__class__.__name__)
    plt.subplot(121)
    covar = kernel_function.function(data, middle_data)
    covar = covar.reshape(-1)
    if index_mode:
        half_label = np.arange(0, data.shape[0])
        middle_label = middle_index
    else:
        half_label = data
        middle_label = middle_data[0, 0]
    plt.plot(half_label, covar, linestyle='-', color='#1f77b4', ms=40)
    plt.xlabel("input, x")
    plt.ylabel("covariance, k(x, {})".format(middle_label))

    plt.subplot(122)
    covar = kernel_function.function(data, data)
    covar = np.flip(covar, 0)
    if index_mode:
        extent = [0, data.shape[0], 0, data.shape[0]]
    else:
        min_data = np.min(data)
        max_data = np.max(data)
        extent = [min_data, max_data, min_data, max_data]
    plt.imshow(
        covar,
        cmap='Blues',
        interpolation='none',
        extent=extent
    )
    cb = plt.colorbar()
    cb.set_label("covariance, k(x, x')")
    plt.xlabel("input, x")
    plt.ylabel("input, x'")
    if file_name is None:
        plt.show()
    else:
        plt.savefig(file_name, bbox_inches='tight')
        plt.close()


def gp_1d(gp, data=None, test_data=None, file_name=None, n_samples=0):
    """
    Plot gp in 1D

    :param gp:
    :type gp:
    :param data:
    :type data:
    :param test_data:
    :type test_data:
    :param file_name:
    :type file_name:
    :param n_samples:
    :type n_samples:
    :return:
    :rtype:
    """
    plt.figure(figsize=(5, 4))

    show_data, min_values, max_values, diff_values = \
        _get_plot_bounds(data, test_data)

    plot_points = np.arange(
        min_values[0],
        max_values[0],
        diff_values[0] / 500.0
    )[:, None]

    mean = gp.mean_function.marginalize_mean(
        plot_points)

    var = gp.covariance_function.marginalize_covariance(
        plot_points, only_diagonal=True)

    sdev = 3.0 * np.sqrt(var)

    plt.clf()
    if 0 < n_samples:
        posterior_samples = gp.sample(plot_points, n_samples=n_samples)
        plt.plot(
            plot_points.flatten().tolist(),
            posterior_samples, color='#1f77b4', linewidth=0.5
        )
    plt.plot(
        plot_points.flatten().tolist(),
        mean.flatten().tolist(), color='#1f77b4', linewidth=2
    )
    ax = plt.gca()
    ax.fill_between(
        plot_points.flatten().tolist(),
        (mean - sdev).flatten().tolist(),
        (mean + sdev).flatten().tolist(),
        color='#aec7e8'
    )
    if show_data is not None:
        plt.plot(
            show_data['X'].flatten().tolist(),
            show_data['Y'].flatten().tolist(), color='#444444',
            linestyle='None', marker='o', markersize=4
        )
    if test_data is not None:
        plt.plot(
            test_data['X'].flatten().tolist(),
            test_data['Y'].flatten().tolist(), color='#d62728',
            linestyle='None', marker='^', markersize=4
        )

    if file_name is None:
        plt.show()
    else:
        plt.savefig(file_name, bbox_inches='tight')
        plt.close()


def gp_2d(gp, data=None, test_data=None, resolution=40, file_name=None):
    """
    Plot gp in 2D

    :param resolution:
    :type resolution:
    :param test_data:
    :type test_data:
    :param file_name:
    :type file_name:
    :param data:
    :type data:
    :param gp:
    :type gp:
    :return:
    :rtype:
    """
    show_data, min_values, max_values, _ = \
        _get_plot_bounds(data, test_data)

    # Init grid
    x_grid = np.linspace(min_values[0], max_values[0], resolution)
    y_grid = np.linspace(min_values[1], max_values[1], resolution)
    x_grid, y_grid = np.meshgrid(x_grid, y_grid)
    plot_points = np.vstack((x_grid.flatten(), y_grid.flatten())).T

    mean = gp.mean_function.marginalize_mean(
        plot_points)

    # Init plot
    fig = plt.figure()
    axis = fig.gca(projection='3d')

    # plot data
    if show_data is not None:
        axis.scatter(
            show_data['X'][:, 0].tolist(),
            show_data['X'][:, 1].tolist(),
            show_data['Y'].tolist(), s=20, color='#444444', marker='o'
        )

    # plot test data
    if test_data is not None:
        axis.scatter(
            test_data['X'][:, 0].tolist(),
            test_data['X'][:, 1].tolist(),
            test_data['Y'].tolist(), s=20, color='#d62728', marker='^'
        )

    # plot GP
    z_points = mean
    z_points = z_points.reshape(resolution, resolution)
    axis.plot_surface(
        x_grid, y_grid,
        z_points, rstride=1,
        cstride=1, cmap='Blues', linewidth=0,
        antialiased=False, alpha=0.3
    )

    axis.zaxis.set_major_locator(LinearLocator(10))
    axis.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    if file_name is None:
        plt.show()
    else:
        plt.savefig(file_name, bbox_inches='tight')
        plt.close()


def _get_plot_bounds(data=None, test_data=None, tail=0.5):
    if test_data is not None:
        focus_data = test_data
    elif data is not None:
        focus_data = data
    else:
        focus_data = {
            'X': np.array([[1.0], [-1.0]])
        }

    points = focus_data['X']
    min_values = np.min(points, axis=0)
    max_values = np.max(points, axis=0)
    diff_values = max_values - min_values
    min_values -= diff_values * tail
    max_values += diff_values * tail
    diff_values = max_values - min_values

    show_data = None
    if data is not None:
        show_i = np.logical_and(
            np.all(min_values < data['X'], axis=1),
            np.all(data['X'] < max_values, axis=1)
        )
        show_data = {
            'X': data['X'][show_i],
            'Y': data['Y'][show_i]
        }

    return show_data, min_values, max_values, diff_values



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .parametrizable import Parametrizable
from .parameter import Parameter


class WithParameters(Parametrizable):
    """

    """
    def __init__(self, name, hyperparameters):
        """

        :param name:
        :type name:
        :param hyperparameters:
        :type hyperparameters:
        """
        self.name = name
        self.hyperparameters = hyperparameters

    def __hash__(self):
        hp_hashes = [hash(self.__class__)]
        for hp in self.hyperparameters:
            if isinstance(hp, WithParameters):
                hp_hashes.append(hash(hp))
            elif isinstance(hp, Parameter):
                if hp.is_array():
                    hp_hashes.extend(hp.current_value)
                else:
                    hp_hashes.append(hp.current_value)
            else:
                raise Exception("Incorrect structure")
        return hash(tuple(hp_hashes))

    def set_param_values(self, params, only_group=None, trans=False):
        """

        :param params:
        :type params:
        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        assert (len(params) == self.get_param_n(
            only_group=only_group)),\
            "length of params is not correct"

        i = 0
        for hyperparameter in self.hyperparameters:
            number_of_params = \
                hyperparameter.get_param_n(only_group=only_group)
            param_slice = slice(i, i + number_of_params)
            hyperparameter.set_param_values(
                params[param_slice],
                only_group=only_group,
                trans=trans
            )
            i += number_of_params

    def get_param_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        params = []
        for hyperparameter in self.hyperparameters:
            params += hyperparameter.get_param_values(
                only_group=only_group,
                trans=trans
            )

        return params

    def get_param_bounds(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        bounds = []
        for hyperparameter in self.hyperparameters:
            bounds += hyperparameter.get_param_bounds(
                only_group=only_group,
                trans=trans
            )

        return bounds

    def get_param_keys(self, recursive=True, only_group=None):
        """

        :param recursive:
        :type recursive:
        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """

        if not recursive:
            return self.name

        params = []

        for hyperparameter in self.hyperparameters:
            name = self.name + self.__class__.__name__
            params += [
                name + "_" + item for item in hyperparameter.get_param_keys(
                    only_group=only_group
                )
            ]

        return params

    def get_param_n(self, only_group=None):
        """

        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """

        n_optimizable = 0

        for hyperparameter in self.hyperparameters:
            n_optimizable += hyperparameter.get_param_n(
                only_group=only_group
            )

        return n_optimizable

    def get_default_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        grid = [
            grid
            for hyperparameter in self.hyperparameters
            for grid in hyperparameter.get_default_values(
                only_group=only_group,
                trans=trans
            )
        ]

        if len(grid) < 1:
            return []

        return grid

    def get_hyperparams(self):
        """

        :return:
        :rtype:
        """

        return self.hyperparameters

    def set_hyperparam(self, name, hyperparam):
        """

        :param name:
        :type name:
        :param hyperparam:
        :type hyperparam:
        :return:
        :rtype:
        """
        i = 0
        n_hps = len(self.hyperparameters)
        while i < n_hps:
            hp_name = self.hyperparameters[i].get_param_keys(
                recursive=False,
                only_group=None
            )
            if name == hp_name:
                self.hyperparameters[i] = hyperparam
                break
            i += 1

    def get_hyperparam(self, name):
        """

        :param name:
        :type name:
        :return:
        :rtype:
        """
        for hyperparameter in self.hyperparameters:
            hp_name = hyperparameter.get_param_keys(
                recursive=False,
                only_group=None
            )
            if name == hp_name:
                return hyperparameter

    def get_param_value(self, name):
        """

        :param name:
        :type name:
        :return:
        :rtype:
        """
        hyperparam = self.get_hyperparam(name)

        value = hyperparam.get_param_values(only_group=None)

        if hyperparam.is_array():
            return value

        return value[0]



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

class Parametrizable(object):
    """

    """

    def set_param_values(self, params, only_group=None, trans=False):
        """

        :param params:
        :type params:
        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_bounds(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_grid(self, limit=1e5, only_group=None, trans=False):
        """

        :param limit:
        :type limit:
        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        def_values = self.get_default_values(only_group=only_group, trans=trans)

        if len(def_values) < 1:
            return []

        all_comb = np.product([len(i) for i in def_values])
        size = min(all_comb, int(limit))

        if all_comb <= 1e5:
            grid = np.array(np.meshgrid(*def_values))
            grid = grid.T.reshape(-1, len(def_values))

            grid = grid[np.random.choice(
                range(len(grid)),
                size=size,
                replace=False
            )]
        else:
            grid = np.array([
                np.random.choice(def_value, size=size)
                for def_value in def_values
            ]).T

        return grid.tolist()

    def get_param_keys(self, recursive=True, only_group=None):
        """

        :param recursive:
        :type recursive:
        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_n(self, only_group=None):
        """

        :param only_group:
        :type only_group:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_default_values(self, only_group=None, trans=False):
        """

        :param only_group:
        :type only_group:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import time

import numpy as np

from .fitting_method import FittingMethod
from ..parameters.parameter import Parameter


class MultiStart(FittingMethod):
    """

    """

    def __init__(self, obj_fun, max_fun_call=500, group=None,
                 nested_fit_method=None):

        self.obj_fun = obj_fun
        self.group = Parameter.OPT_GROUP
        if group is not None:
            self.group = group
        self.max_fun_call = max_fun_call
        super(MultiStart, self).__init__(nested_fit_method)

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """
        if budget is None or self.max_fun_call < budget:
            max_fun_call = self.max_fun_call
        else:
            max_fun_call = budget

        start = time.time()

        best_state = {
            'name': self.__class__.__name__,
            'fun_calls': 0,
            'improvements': 0,
            'restarts': 0,
            'time': 0,
            'value': np.inf
        }
        best_state.update(self.save_state(model))
        initial_state = self.save_state(model)

        param_grid = model.get_grid(
            limit=max_fun_call,
            only_group=self.group,
            trans=True
        )
        current_opt_params = np.array(model.get_param_values(
            only_group=self.group,
            trans=True
        ))
        best_opt_params = current_opt_params

        while best_state['fun_calls'] < max_fun_call:
            # run optimization
            value = np.inf
            self.load_state(model, initial_state)
            model.set_param_values(
                current_opt_params,
                only_group=self.group,
                trans=True
            )
            nested_log = None
            try:
                if self.nested_fit_method is not None:
                    nested_log = self.nested_fit_method.fit(
                        model,
                        folds,
                        budget=(max_fun_call - best_state['fun_calls'] - 1),
                        verbose=verbose
                    )
                    best_state['fun_calls'] += (nested_log['fun_calls'])
                value = self.obj_fun(
                    model=model,
                    folds=folds,
                    grad_needed=False
                )
            except (AssertionError, np.linalg.linalg.LinAlgError) as ex:
                if verbose:
                    print(ex)

            best_state['fun_calls'] += 1
            best_state['restarts'] += 1
            if value < best_state['value']:
                best_state['improvements'] += 1
                best_state['value'] = value
                best_state['state'] = model.get_param_values(
                    only_group=self.group,
                    trans=False
                )
                best_state['nested'] = nested_log
                best_opt_params = model.get_param_values(
                    only_group=self.group,
                    trans=True
                )

            if len(param_grid) < 1:
                break

            if np.random.uniform() < 0.1:
                current_opt_params = np.array(
                    param_grid[np.random.randint(len(param_grid))]
                )
                jitter_sd = 10.
            else:
                current_opt_params = np.array(best_opt_params)
                jitter_sd = 1.
            current_opt_params += np.random.normal(
                loc=0.0,
                scale=jitter_sd,
                size=len(current_opt_params)
            )

        end = time.time()

        assert best_state['improvements'], "No params were found"

        self.load_state(model, best_state)

        best_state['time'] = end - start

        return best_state

    def save_state(self, model):
        """

        :param model:
        :type model:
        :return:
        :rtype:
        """
        state = self.nested_save_state(model)
        state['state'] = model.get_param_values(
            only_group=self.group,
            trans=False
        )

        return state

    def load_state(self, model, state):
        """

        :param model:
        :type model:
        :param state:
        :type state:
        :return:
        :rtype:
        """
        model.set_param_values(
            state['state'],
            only_group=self.group,
            trans=False
        )
        self.nested_load_state(model, state)



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .metric import Metric


class LML(Metric):
    """

    """

    @staticmethod
    def measure(model, train_set, test_set=None, grad_needed=False):
        """

        :param model:
        :type model:
        :param train_set:
        :type train_set:
        :param test_set:
        :type test_set:
        :param grad_needed:
        :type grad_needed:
        :return:
        :rtype:
        """

        if test_set:
            current_model = model.get_posterior(train_set)
            data_set = test_set
        else:
            current_model = model
            data_set = train_set

        result = current_model.get_log_likelihood(
            data_set,
            grad_needed=grad_needed
        )

        if grad_needed:
            prior_lml, gradient = result
            prior_lml = -float(prior_lml)
            gradient = -np.array(gradient)
            return prior_lml, gradient

        prior_lml = result
        prior_lml = -float(prior_lml)
        return prior_lml



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class Validation(object):
    """

    """
    def get_folds(self, data_set):
        """

        :param data_set:
        :type data_set:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    @staticmethod
    def split_data(data_set, n_reps, stratified, shuffle, folds_n, fold_len):
        """

        :param data_set:
        :type data_set:
        :param n_reps:
        :type n_reps:
        :param stratified:
        :type stratified:
        :param shuffle:
        :type shuffle:
        :param folds_n:
        :type folds_n:
        :param fold_len:
        :type fold_len:
        :return:
        :rtype:
        """

        folds = []

        if stratified:
            groups = [
                np.argwhere(data_set['Y'] == y)[:, 0]
                for y in np.unique(data_set['Y'])
            ]
        else:
            groups = [np.arange(data_set['X'].shape[0])]

        for _ in range(n_reps):
            if shuffle:
                groups = [
                    group[np.random.permutation(len(group))]
                    for group in groups
                ]
            for fold_i in range(folds_n):
                train_selection = np.array([], dtype=np.int64)
                test_selection = np.array([], dtype=np.int64)
                for group in groups:
                    n = len(group)
                    if fold_len is not None:
                        rel_fold_len = \
                            int(np.ceil(float(n) * fold_len))
                    else:
                        rel_fold_len = \
                            int(np.ceil(float(n) / folds_n))
                    cut1 = max(n - (rel_fold_len * (fold_i + 1)), 0)
                    cut2 = n - (rel_fold_len * fold_i)
                    train_slice1 = slice(None, cut1, None)
                    test_slice = slice(cut1, cut2, None)
                    train_slice2 = slice(cut2, None, None)
                    train_selection = np.concatenate((
                        train_selection,
                        group[train_slice1],
                        group[train_slice2]
                    ))
                    test_selection = np.concatenate((
                        test_selection,
                        group[test_slice]
                    ))
                train_set = {
                    name: element[train_selection, :]
                    for name, element in data_set.items()
                }
                test_set = None
                if len(test_selection) > 0:
                    test_set = {
                        name: element[test_selection, :]
                        for name, element in data_set.items()
                    }
                folds.append((train_set, test_set))

        return folds
